# dataset.py
"""Class Dataset: A dictionary of DataElements, which in turn can have a Sequence of Datasets.

Overview of Dicom object model:
------------------------------
Dataset(derived class of Python's dict class)
    ---> contains DataElement instances (DataElement is a class with tag, VR, value)
             ==> the value can be a Sequence instance (Sequence is derived from Python's list),
                            or just a regular value like a number, string, etc.,
                            or a list of regular values, e.g. a 3d coordinate
                  --> Sequence's are a list of Datasets
   
"""
#
# Copyright 2008, Darcy Mason
# This file is part of pydicom.
#
# pydicom is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pydicom is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (license.txt) for more details

from sys import byteorder
sys_isLittleEndian = (byteorder == 'little')
from dicom.datadict import DicomDictionary, dictionaryVR
from dicom.datadict import TagForName, AllNamesForTag
from dicom.tag import Tag
from dicom.dataelem import DataElement
import dicom # for WriteFile
import dicom.charset

haveNumpy = True

try:
    import numpy
except:
    haveNumpy = False

class Dataset(dict):
    """A Dataset is a collection (dictionary) of Dicom DataElement instances.
    
    Example of two ways to retrieve or set values:
    (1) dataset[0x10, 0x10].value --> patient's name
    (2) dataset.PatientsName --> patient's name
    
    Example (2) is referred to as *Named tags* in this documentation.
    PatientsName is not actually a member of the object, but unknown member
    requests are checked against the dicom dictionary. If the name matches a
    DicomDictionary descriptive string, the corresponding tag is used
    to look up or set the Data Element's value.
    
    Class Data
    ----------
    indentChars -- for string display, the characters used to indent for
            nested Data Elements (e.g. sequence items). Default is '   '.
    
    NumpyPixelFormats -- if NumPy module is available, map bits allocated
            to the NumPy typecode.
    """
    indentChars = "   "
    # Map image_dataset.BitsAllocated to NumPy typecode
    if haveNumpy:
        NumpyPixelFormats = {8: numpy.uint8, 16:numpy.int16, 32:numpy.int32}
                
    def Add(self, data_element):
        """Equivalent to dataset[data_element.tag] = data_element."""
        self[data_element.tag] = data_element
    
    def AddNew(self, tag, VR, value):
        """Create a new DataElement instance and add it to this Dataset."""
        data_element = DataElement(tag, VR, value)
        self[data_element.tag] = data_element   # use data_element.tag since DataElement verified it
    
    def attribute(self, name):
        """Deprecated -- use Dataset.data_element()"""
        import warnings
        warnings.warn("Dataset.attribute() is deprecated and will be removed in pydicom 1.0. Use Dataset.data_element() instead", DeprecationWarning)
        return self.data_element(name)
    def data_element(self, name):
        """Return the full data_element instance for the given descriptive name.
        
        When using *named tags*, only the value is returned. If you want the
        whole data_element object, for example to change the data_element.VR,
        call this function with the name and the data_element instance is returned."""
        tag = TagForName(name)
        if tag:
            return self[tag]
        return None
    
    def __contains__(self, name):
        """Extend dict.__contains__() to handle *named tags*.
        
        This is called for code like: ``if 'SliceLocation' in dataset``.
        
        """
        if self._isString(name):
            tag = TagForName(name)
        else:
            try:
                tag = Tag(name)
            except:
                return 0
        if tag:
            return dict.__contains__(self, tag)
        else:
            return dict.__contains__(self, name) # will no doubt raise an exception
    
    def decode(self):
        """Apply character set decoding to all data elements.
        
        See DICOM PS3.5-2008 6.1.1.
        """
        # Find specific character set. 'ISO_IR 6' is default
        # May be multi-valued, but let dicom.charset handle all logic on that
        dicom_character_set = self.get('SpecificCharacterSet', "ISO_IR 6")

        # shortcut to the decode function in dicom.charset
        decode_data_element = dicom.charset.decode

        # sub-function callback for walk(), to decode the chr strings if necessary
        # this simply calls the dicom.charset.decode function
        def decode_callback(ds, data_element):
            decode_data_element(data_element, dicom_character_set)
        # Use the walk function to go through all elements in the dataset and convert them
        self.walk(decode_callback)
    
    def __dir__(self):
        """___dir__ is used in python >= 2.6 to give a list of attributes
        available in an object, for example used in auto-completion in editors
        or command-line environments.
        """
        return self.dir()
    
    def dir(self, *filters):
        """Return a list of some or all data_element names, in alphabetical order.
        
        Intended mainly for use in interactive Python sessions.
        
        filters -- 0 or more string arguments to the function
                if none provided, dir() returns all data_element names in this Dataset.
                Else dir() will return only those items with one of the strings
                somewhere in the name (case insensitive).
        
        """
        allnames = []
        for tag, data_element in self.items():
            allnames.extend(AllNamesForTag(tag))
        allnames = [x for x in allnames if x]  # remove blanks - tags without valid names (e.g. private tags)
        # Store found names in a dict, so duplicate names appear only once
        matches = {}
        for filter in filters:
            filter = filter.lower()
            match = [x for x in allnames if x.lower().find(filter) != -1]
            matches.update(dict([(x,1) for x in match]))
        if filters:
            names = matches.keys()
            names.sort()
            return names
        else:
            allnames.sort()
            return allnames

    def file_metadata(self):
        """Return a Dataset holding only meta information (group 2).
        
        Only makes sense if this dataset is a whole file dataset.
        
        """
        return GroupDataset(2)
        
    def get(self, key, default=None):
        """Extend dict.get() to handle *named tags*."""
        if self._isString(key):
            try:
                return getattr(self, key)
            except AttributeError:
                return default
        else: # is not a string, probably is a tag -> hand off to underlying dict
            return dict.get(self, key, default)
    def __getattr__(self, name):
        """Intercept requests for unknown Dataset python-attribute names.
        
        If the name matches a Dicom dictionary string (without blanks etc),
        then return the value for the data_element with the corresponding tag.
        
        """
        # __getattr__ only called if instance cannot find name in self.__dict__
        # So, if name is not a dicom string, then is an error
        tag = TagForName(name)
        if not tag or tag not in self:
            raise AttributeError, "Dataset does not have attribute '%s'." % name
        else:  # do have that dicom data_element
            return self[tag].value
    def __getitem__(self, key):
        """Operator for dataset[key] request."""
        return dict.__getitem__(self, Tag(key))
    def GroupDataset(self, group):
        """Return a Dataset containing only data_elements of a certain group.
        
        group -- the group part of a dicom (group, element) tag.
        
        """
        ds = Dataset()
        ds.update(dict(
            [(tag,data_element) for tag,data_element in self.items() if tag.group==group]
                      ))
        return ds
    def has_key(self, key):
        """Extend dict.has_key() to handle *named tags*."""
        return self.__contains__(key)
    
    # isBigEndian property
    def _getBigEndian(self):
        return not self.isLittleEndian
    def _setBigEndian(self, value):
        self.isLittleEndian = not value
    isBigEndian = property(_getBigEndian, _setBigEndian)
    
    def _isString(self, name):
        """Return True if name is a string."""
        try:
            name + ""
        except TypeError:
            return 0
        else:
            return 1

    def __iter__(self):
        """Method to iterate through the dataset, returning data_elements.
        e.g.:
        for data_element in dataset:
            do_something...
        The data_elements are returned in DICOM order, 
        i.e. in increasing order by tag value.
        Sequence items are returned as a single data_element; it is up to the
           calling code to recurse into the Sequence items if desired
        """
        # Note this is different than the underlying dict class, 
        #        which returns the key of the key:value mapping.
        #   Here the value is returned (but data_element.tag has the key)
        taglist = self.keys()
        taglist.sort()
        for tag in taglist:
            yield self[tag]
            
            
    def _PixelDataNumpy(self):
        """Return a NumPy array of the pixel data.
        
        NumPy is the most recent numerical package for python. It is used if available.
        
        Raise TypeError if no pixel data in this dataset.
        Raise ImportError if cannot import numpy.
        
        """
        if not 'PixelData' in self:
            raise TypeError, "No pixel data found in this dataset."

        if not haveNumpy:
            msg = "The Numpy package is required to use PixelArray and numpy could not be imported.\n"
            raise ImportError, msg

        # determine the type used for the array
        need_byteswap = (self.isLittleEndian != sys_isLittleEndian)

        if haveNumpy:
            if self.BitsAllocated not in self.NumpyPixelFormats:
                raise NotImplementedError, "Do not have NumPy dtype for BitsAllocated=%d, please update Dataset.NumpyPixelFormats" % self.BitsAllocated
            format = self.NumpyPixelFormats[self.BitsAllocated]
            arr = numpy.fromstring(self.PixelData, format)
            # XXX byte swap - may later handle this in ReadFile!!?
            if need_byteswap:
                arr.byteswap(True)  # True = swap in-place, don't make a new copy
            # Note the following reshape operations return a new *view* onto arr, but don't copy the data
            if 'NumberofFrames' in self and self.NumberofFrames > 1:
                if self.SamplesperPixel > 1:
                    arr = arr.reshape(self.SamplesperPixel, self.NumberofFrames, self.Rows, self.Columns)
                else:
                    arr = arr.reshape(self.NumberofFrames, self.Rows, self.Columns)
            else:
                if self.SamplesperPixel > 1:
                    if self.BitsAllocated == 8:
                        arr = arr.reshape(self.SamplesperPixel, self.Rows, self.Columns)
                    else:
                        raise NotImplementedError, "This code only handles SamplesPerPixel > 1 if Bits Allocated = 8"
                else:
                    arr = arr.reshape(self.Rows, self.Columns)
        return arr
    
    # PixelArray property
    def _getPixelArray(self):
        if self.TransferSyntaxUID not in ['1.2.840.10008.1.2',        
        # Check if pixel data is in a form we know how to make into an array
                                          '1.2.840.10008.1.2.1',
                                          '1.2.840.10008.1.2.1.99',
                                          '1.2.840.10008.1.2.2']: 
            raise NotImplementedError, "Pixel Data is compressed in a format pydicom does not yet handle. Cannot return array"

        # Check if already have converted to a NumPy array
        # Also check if self.PixelData has changed. If so, get new NumPy array
        alreadyHave = True
        if not hasattr(self, "_PixelArray"):
            alreadyHave = False
        elif self._pixel_id != id(self.PixelData):
            alreadyHave = False
        if not alreadyHave:    
            self._PixelArray = self._PixelDataNumpy()
            self._pixel_id = id(self.PixelData) # is this guaranteed to work if memory is re-used??
        return self._PixelArray
    PixelArray = property(_getPixelArray)
    
    def _PrettyStr(self, indent=0, topLevelOnly=False):
        """Return a string of the data_elements in this dataset, with indented levels.
        
        This private method is called by the __str__() method 
        for handling print statements or str(dataset), and the __repr__() method.
        It is also used by top(), which is the reason for the topLevelOnly flag.
        This function recurses, with increasing indentation levels.
        
        """
        strings = []
        indentStr = self.indentChars * indent
        nextIndentStr = self.indentChars *(indent+1)
        for data_element in self:
            if data_element.VR == "SQ":   # a sequence
                strings.append(indentStr + str(data_element.tag) + "  %s   %i item(s) ---- " % ( data_element.description(),len(data_element.value)))
                if not topLevelOnly:
                    for dataset in data_element.value:
                        strings.append(dataset._PrettyStr(indent+1))
                        strings.append(nextIndentStr + "---------")
            else:
                strings.append(indentStr + repr(data_element))
        return "\n".join(strings)
        
    def RemovePrivateTags(self):
        """Remove all Dicom private tags in this dataset and those contained within."""
        def RemoveCallback(dataset, data_element):
            """Internal method to use as callback to walk() method."""
            if data_element.tag.isPrivate:
                # can't del self[tag] - won't be right dataset on recursion
                del dataset[data_element.tag]  
        self.walk(RemoveCallback)

    def SaveAs(self, filename, WriteLikeOriginal=True):
        """Write the dataset to a file.
        
        filename -- full path and filename to save the file to
        WriteLikeOriginal -- see dicom.filewrite.WriteFile for info on this parameter.
        """
        dicom.WriteFile(filename, self, WriteLikeOriginal)
    
    def __setattr__(self, name, value):
        """Intercept any attempts to set a value for an instance attribute.
        
        If name is a dicom descriptive string (cleaned with CleanName),
        then set the corresponding tag and data_element.
        Else, set an instance (python) attribute as any other class would do.
        
        """
        tag = TagForName(name)
        if tag:  # successfully mapped name to a tag
            if tag not in self:  # don't have this tag yet->create the data_element instance
                VR = dictionaryVR(tag)
                data_element = DataElement(tag, VR, value)
            else:  # already have this data_element, just changing its value
                data_element = self[tag]
                data_element.value = value
            # Now have data_element - store it in this dict
            self[tag] = data_element
        else:  # name not in dicom dictionary - setting a non-dicom instance attribute
            # XXX note if user mis-spells a dicom data_element - no error!!!
            self.__dict__[name] = value  

    def __setitem__(self, key, value):
        """Operator for dataset[key]=value."""
        try:
            x = value.VR, value.tag, value.value  # check is an DataElement by its contents
        except AttributeError:
            raise TypeError, "Dataset contents must be DataElement instances.\n" + \
                  "To set a data_element value use data_element.value=val"
        if key != value.tag:
            raise ValueError, "data_element.tag must match the dictionary key"
        # Everything is okay - use base class (dict) to store the dicom DataElement
        dict.__setitem__(self, value.tag, value)

    def __str__(self):
        """Handle str(dataset)."""
        return self._PrettyStr()
        
    def top(self):
        """Show the DICOM tags, but only the top level; do not recurse into Sequences"""
        return self._PrettyStr(topLevelOnly=True)

    def trait_names(self):
        """Return a list of valid name for auto-completion code
        Used in IPython, so that data element names can be found
        and offered for autocompletion on the IPython command line
        """
        return self.dir() # does not list underlying dict properties and methods
        
    def update(self, dictionary):
        """Extend dict.update() to handle *named tags*."""
        for key, value in dictionary.items():
            if self._isString(key):
                setattr(self, key, value)
            else:
                self[Tag(key)] = value
                
    def walk(self, callback):
        """Call the given function for all dataset data_elements (recurses).
        
        Go through all data_elements, recursing into sequences and their datasets,
        calling the callback function at each data_element (including the SQ data_element).
        This can be used to perform an operation on certain types of data_elements.
        For example, RemovePrivateTags() finds all private tags and deletes them.

        callback -- a callable method which takes two arguments: a dataset, and
                    a data_element belonging to that dataset.
        
        DataElements will come back in dicom order (by increasing tag number
        within their dataset)
        
        """
        taglist = self.keys()
        taglist.sort()
        for tag in taglist:
            data_element = self[tag]
            callback(self, data_element)  # self = this Dataset
            # 'tag in self' below needed in case data_element was deleted in callback
            if tag in self and data_element.VR == "SQ":  
                sequence = data_element.value
                for dataset in sequence:
                    dataset.walk(callback)

    __repr__ = __str__
